/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of PeerfactSim.KOM.
 * 
 * PeerfactSim.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tud.kom.p2psim.impl.topology.movement.modularosm.groups.groupencounter;

import de.tud.kom.p2psim.api.scenario.ConfigurationException;
import de.tud.kom.p2psim.impl.topology.movement.modularosm.SocialGroupMovementModel;
import de.tud.kom.p2psim.impl.topology.movement.modularosm.groups.SocialMovementGroup;
import de.tudarmstadt.maki.simonstrator.api.util.XMLConfigurableConstructor;

/**
 * This class checks if there is any pair of groups which encounter each other. Randomly decides
 * if the groups merge fully, dissolve or do nothing after the encounter.
 * 
 * @author Marcel Verst
 * @version 1.0, 22.11.2018
 */
public class MergeOrDissolve extends AbstractGroupEncounter {
	//--- VARIABLES TO BE DECLARED IN CONFIG ---
	private double probabilityToMergeFully;
	private double probabilityToDissolve;
	//------------------------------------------

	@XMLConfigurableConstructor({"enableGroupEncounters","groupEncounterMeetingDistance","groupReencounterWaitTime"})
	public MergeOrDissolve(boolean enableGroupEncounters,
			double groupEncounterMeetingDistance,
			long groupReencounterWaitTime) {
		super(enableGroupEncounters, groupEncounterMeetingDistance,
				groupReencounterWaitTime);		
	}
	
	@Override
	public void initialize(SocialGroupMovementModel movementModel) {
		super.initialize(movementModel);
		checkConfiguration();
	}
	
	/**
	 * {@inheritDoc}
	 * Applies dependent on probability values if the groups merge fully, dissolve or do nothing.
	 */
	@Override
	public void handleGroupEncounter(SocialMovementGroup group1, SocialMovementGroup group2) {
		int decision = decide();

		if(decision == DECISION_MERGE_FULL) {
			mergeGroupsFully(group1, group2);
		}
		else if(decision == DECISION_DISSOLVE) {
			dissolveGroups(group1, group2);
		}
	}
	
	/**
	 * Decider to set the function which a leader should call if its group is in meeting range with another group.
	 * 
	 * @return int
	 */
	protected int decide() {
		
		double decider = rand.nextDouble();

		double maxMergeFully = probabilityToMergeFully;
		double maxDissolve = probabilityToMergeFully + probabilityToDissolve;

		// mergeFully
		if(decider >= 0.0 && decider < maxMergeFully) {
			return DECISION_MERGE_FULL;
		}
		// dissolve
		else if(decider >= maxMergeFully && decider < maxDissolve) {
			return DECISION_DISSOLVE;
		}
		else {
			return DECISION_NONE;
		}
	}
	
	public void setProbabilityToMergeFully(double probabilityToMergeFully) {
		if (probabilityToMergeFully < 0 || probabilityToMergeFully > 1.0) {
			throw new ConfigurationException(
					"probabilityToMergeFully should be between 0.0 and 1.0!");
		}
		this.probabilityToMergeFully = probabilityToMergeFully;
	}
	
	public void setProbabilityToDissolve(double probabilityToDissolve) {
		if (probabilityToDissolve < 0 || probabilityToDissolve > 1.0) {
			throw new ConfigurationException(
					"probabilityToDissolve should be between 0.0 and 1.0!");
		}
		this.probabilityToDissolve = probabilityToDissolve;
	}
	
	/**
	 * Checks for the correct configuration.
	 * 
	 * @author Marcel Verst
	 */
	private void checkConfiguration() {
		if(probabilityToMergeFully + probabilityToDissolve > 1.0) {
			throw new ConfigurationException(
					"FullyDissolve Group Encounter: probabilityToMergeFully + probabilityToDissolve must not exceed 1.0!");
		}
	}
}
