/*
 * Copyright (c) 2005-2010 KOM – Multimedia Communications Lab
 *
 * This file is part of PeerfactSim.KOM.
 * 
 * PeerfactSim.KOM is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * any later version.
 * 
 * PeerfactSim.KOM is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with PeerfactSim.KOM.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

package de.tud.kom.p2psim.impl.util.db.metric;

import jakarta.persistence.Column;
import jakarta.persistence.Entity;
import jakarta.persistence.GeneratedValue;
import jakarta.persistence.GenerationType;
import jakarta.persistence.Id;
import jakarta.persistence.Index;
import jakarta.persistence.JoinColumn;
import jakarta.persistence.ManyToOne;
import jakarta.persistence.Table;


/**
 * A single measurement, but with an assigned location.
 *
 * @author Bjoern Richerzhagen
 */
@Entity
@Table(name = "measurements_spatial", indexes = {
		@Index(columnList = "time", name = "time"),
		@Index(columnList = "hostMetricId", name = "hostMetricId") })
public class MeasurementSpatial {

	/**
	 * A unique Id, will be set by the database
	 */
	@Id
	@GeneratedValue(strategy = GenerationType.IDENTITY)
	private int id;

	/**
	 * The simulation time of this measurement
	 */
	@SuppressWarnings("unused")
	private long time;

	/**
	 * The value for this measurement
	 */
	@Column(nullable = true)
	private Double value;
	
	@Column(nullable = true, name = "[locationX]")
	private Integer locationX;
	
	@Column(nullable = true, name = "[locationY]")
	private Integer locationY;

	/**
	 * The mapping Object of this measurement to the {@link Metric}-Object,
	 * which describes this metric.
	 */
	@ManyToOne
	@JoinColumn(name = "hostMetricId")
	private HostMetric hostMetric;

	/**
	 * Creates a {@link MeasurementSingle}-Object with the given parameters. If
	 * value has the value infinity or NaN, then will be set this value to null.
	 *
	 * @param time
	 *            The simulation time as date to this measurement
	 * @param value
	 *            The value for this measurement
	 * @param statistic
	 *            The reference to the {@link Metric}-Object, which describes
	 *            this metric. Is used for the mapping.
	 */
	public MeasurementSpatial(long time, Double value, HostMetric hostMetric, Integer locationX, Integer locationY) {
		super();
		this.time = time;
		this.hostMetric = hostMetric;

		// check for infinity or NaN
		this.value = checkForSpecialNumbers(value);
		this.locationX = locationX;
		this.locationY = locationY;
	}

	/**
	 * Check for special numbers, like infinity or NaN. If the given value is
	 * equals this numbers then will be return null, otherwise will be returned
	 * the given value.
	 *
	 * @param value
	 *            The value, which should be checked.
	 * @return The value or null, if it is a special number.
	 */
	private Double checkForSpecialNumbers(Double value) {
		if (value == null)
			return value;
		if (value.equals(Double.NEGATIVE_INFINITY)
				|| value.equals(Double.POSITIVE_INFINITY)
				|| value.equals(Double.NaN)) {
			return null;
		} else {
			return value;
		}
	}
	
}
